#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <mpi.h>

static int const NUM_NUMBERS_PER_PROCESS = 10;
static int const ROOT_PROCESS = 0;
static int const MPI_MESSAGE_HELLO_WORLD_TAG = 50;

static int computeMax(int const arr[], int num)
{
  int i, max;
  max = arr[0];
  for (i = 1; i < num; ++i)
  {
    if (max < arr[i])
    {
      max = arr[i];
    }
  }
  return max;
}



int main(int argc, char * argv[])
{
  int myProcessNo;
  int numProcesses;
  int * myNumbers;
  int * allNumbers;
  int * partialResults;
  int maxNumber;
  int printNumbers = 0;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &myProcessNo);
  MPI_Comm_size(MPI_COMM_WORLD, &numProcesses);

  if (argc > 1)
  {
    printNumbers = 1;
  }

  myNumbers = (int *)malloc(sizeof(int) * NUM_NUMBERS_PER_PROCESS);
  if (myNumbers == NULL)
  {
    fprintf(stderr, "MALLOC FAILED IN PROCESS %d!\n", myProcessNo);
    MPI_Finalize();
    return 1;
  }

  if (myProcessNo == ROOT_PROCESS)
  {
    int i;
    allNumbers = (int *)malloc(sizeof(int) * NUM_NUMBERS_PER_PROCESS * numProcesses);
    if (allNumbers == NULL)
    {
      free(myNumbers);
      fprintf(stderr, "MALLOC FAILED IN PROCESS %d!\n", myProcessNo);
      MPI_Finalize();
      return 1;
    }
    partialResults = (int *)malloc(sizeof(int) * numProcesses);
    if (partialResults == NULL)
    {
      free(myNumbers);
      free(allNumbers);
      fprintf(stderr, "MALLOC FAILED IN PROCESS %d!\n", myProcessNo);
      MPI_Finalize();
      return 1;
    }
    for (i = 0; i < NUM_NUMBERS_PER_PROCESS * numProcesses; ++i)
    {
      allNumbers[i] = rand() % (NUM_NUMBERS_PER_PROCESS * numProcesses);
    }
    if (printNumbers)
    {
      printf("The numbers:");
      for (i = 0; i < NUM_NUMBERS_PER_PROCESS * numProcesses; ++i)
      {
        printf(" %d", allNumbers[i]);
      }
      printf(".\n");
    }
  }
  else
  {
    allNumbers = NULL;
    partialResults = NULL;
  }
  
  MPI_Scatter(
      allNumbers,
      NUM_NUMBERS_PER_PROCESS,
      MPI_INT,
      myNumbers,
      NUM_NUMBERS_PER_PROCESS,
      MPI_INT,
      ROOT_PROCESS,
      MPI_COMM_WORLD
  );
  
  maxNumber = computeMax(myNumbers, NUM_NUMBERS_PER_PROCESS);
  
  if (printNumbers)
  {
    printf("The result of process %d is %d.\n", myProcessNo, maxNumber);
  }
  
  MPI_Gather(
      &maxNumber,
      1 /* just one number */,
      MPI_INT,
      partialResults,
      1 /* one number per process */,
      MPI_INT,
      ROOT_PROCESS,
      MPI_COMM_WORLD
  );
  
  if (printNumbers)
  {
    /* The barrier is here only to ensure that */
    /* the final maximum is printed AFTER      */
    /* the partial result of each process      */
    MPI_Barrier(MPI_COMM_WORLD);
  }
  
  if (myProcessNo == ROOT_PROCESS)
  {
    maxNumber = computeMax(partialResults, numProcesses);
    if (printNumbers)
    {
      printf("The final result: %d.\n", maxNumber);
    }
    free(partialResults);
    free(allNumbers);
  }

  free(myNumbers);

  MPI_Finalize();

  return 0;
}
